import logging
from concurrent.futures import ProcessPoolExecutor, as_completed
from pathlib import Path

from git import Repo
from pydantic import (
    BaseModel,
)

from mcpunk.file_chunk import Chunk, ChunkCategory
from mcpunk.file_chunkers import (
    BaseChunker,
    MarkdownChunker,
    PythonChunker,
    VueChunker,
    WholeFileChunker,
)

ALL_CHUNKERS: list[type[BaseChunker]] = [
    PythonChunker,
    MarkdownChunker,
    VueChunker,
    # Want the WholeFileChunker to be last as it's more of a "fallback" chunker
    WholeFileChunker,
]

logger = logging.getLogger(__name__)


class NoGitRepoError(Exception):
    pass


class File(BaseModel):
    chunks: list[Chunk]
    abs_path: Path
    contents: str
    ext: str  # File extension

    @classmethod
    def from_file_contents(
        cls,
        source_code: str,
        file_path: Path,
    ) -> "File":
        """Extract all callables, calls and imports from the given source code file."""
        chunks: list[Chunk] = []

        # Try all eligible chunkers in order until one of them doesn't crash.
        for chunker in ALL_CHUNKERS:
            if chunker.can_chunk(source_code, file_path):
                try:
                    chunks = chunker(source_code, file_path).chunk_file()
                    break
                except Exception:
                    logger.exception(f"Error chunking file {file_path} with {chunker}")
        return File(
            chunks=chunks,
            abs_path=file_path.absolute(),
            contents=source_code,
            ext=file_path.suffix,
        )

    def chunks_of_type(self, chunk_type: ChunkCategory) -> list[Chunk]:
        return [c for c in self.chunks if c.category == chunk_type]


class Project(BaseModel):
    root: Path
    files: list[File]

    @property
    def git_repo(self) -> Repo:
        return _git_repo(self.root)

    @classmethod
    def from_files(cls, root: Path, files: list[Path], max_workers: int | None = None) -> "Project":
        files_analysed: list[File] = []

        with ProcessPoolExecutor(max_workers=max_workers) as executor:
            # Submit all file analysis tasks
            future_to_file = {
                executor.submit(_analyze_file, file_path): file_path for file_path in files
            }

            for future in as_completed(future_to_file):
                file_path = future_to_file[future]
                try:
                    result = future.result()
                    if result is not None:
                        files_analysed.append(result)
                except Exception:
                    logger.exception(f"File {file_path} generated an exception")

        project = Project(
            root=root,
            files=files_analysed,
        )
        return project

    @classmethod
    def from_root_dir(cls, root: Path, max_workers: int | None = None) -> "Project":
        if not root.exists():
            raise ValueError(f"Root directory {root} does not exist")

        repo: Repo | None
        try:
            repo = _git_repo(root)
        except NoGitRepoError:
            repo = None

        files: list[Path] = []
        if repo is not None:
            rel_paths = repo.git.ls_files().splitlines()
            files.extend(root / rel_path for rel_path in rel_paths)
        else:
            # Exclude specific top-level directories
            # TODO: make this configurable
            ignore_dirs = {".venv", "build", ".git", "__pycache__"}  # customize this set

            for path in root.iterdir():
                if path.is_dir() and path.name not in ignore_dirs:
                    files.extend(path.glob("**/*"))

            # Don't forget files in the root directory itself
            files.extend(root.glob("*"))

        files = [file for file in files if file.is_file()]
        return Project.from_files(root, files, max_workers=max_workers)


def _analyze_file(file_path: Path) -> File | None:
    try:
        if not file_path.exists():
            logger.warning(f"File {file_path} does not exist")
            return None
        if not file_path.is_file():
            logger.warning(f"File {file_path} is not a file")
            return None

        return File.from_file_contents(file_path.read_text(), file_path)
    except Exception:
        logger.exception(f"Error processing file {file_path}")
        return None


def _git_repo(root: Path) -> Repo:
    if not (root / ".git").exists():
        raise NoGitRepoError(f"No git repo found at {root}")
    return Repo(root / ".git")



from datetime import timedelta
from pathlib import Path
from typing import Annotated, Literal

from pydantic import AfterValidator
from pydantic_settings import BaseSettings, SettingsConfigDict


def _post_fiddle_path(p: Path) -> Path:
    return p.expanduser().absolute()


class Settings(BaseSettings):
    # SQLite database path
    db_path: Annotated[
        Path,
        AfterValidator(_post_fiddle_path),
    ] = Path("~/.mcpunk/db.sqlite")

    # I believe that MCP clients should not look at stderr, but it seems some do
    # which completely messes with things. Suggest leaving this off and relying
    # on the log *file* instead.
    enable_stderr_logging: bool = False

    enable_log_file: bool = True
    log_file: Annotated[
        Path,
        AfterValidator(_post_fiddle_path),
    ] = Path("~/.mcpunk/mcpunk.log")
    log_level: Literal["DEBUG", "INFO", "WARNING", "ERROR", "FATAL", "CRITICAL"] = "INFO"

    # Default indentation when serialising to JSON in response.
    default_response_indent: int | Literal["no_indent"] = 2
    # Whether to include the number of characters in the response in the response.
    # So like `[DEBUG INFO: Response is 1234 chars]` prefixed to the response.
    include_chars_in_response: bool = True
    # Maximum number of characters in the response. If the response is longer than this
    # then an error will be returned to the caller. This is handy to avoid blowing
    # your context.
    default_response_max_chars: int = 20_000
    # Same as `default_response_max_chars` but for the tool that returns a git diff.
    # Generally, git diffs are a bit larger than e.g. a function so nice to have it a
    # bit larger.
    default_git_diff_response_max_chars: int = 50_000

    # A task which is in the "doing" state for longer than this duration
    # will become available again for pickup.
    task_queue_visibility_timeout_seconds: int = 300

    @property
    def task_queue_visibility_timeout(self) -> timedelta:
        return timedelta(seconds=self.task_queue_visibility_timeout_seconds)

    model_config = SettingsConfigDict(
        env_prefix="MCPUNK_",
        validate_default=True,
        frozen=True,
    )



from abc import abstractmethod
from pathlib import Path

from bs4 import BeautifulSoup

from mcpunk.file_chunk import Chunk, ChunkCategory
from mcpunk.python_file_analysis import Callable, extract_imports, extract_module_statements


class BaseChunker:
    """Base class for file chunkers."""

    def __init__(self, source_code: str, file_path: Path) -> None:
        self.source_code = source_code
        self.file_path = file_path

    @staticmethod
    @abstractmethod
    def can_chunk(source_code: str, file_path: Path) -> bool:
        """Return True if the file can likely be chunked by this class.

        This should be a very quick cheap check, quite possibly just using the file
        extension. Do not assume that the file exists on disk.

        Users of file chunks should handle gracefully the case where this returns
        True but `chunk_file` fails. For example, the file may appear to be Python
        but could contain invalid syntax.
        """
        raise NotImplementedError

    @abstractmethod
    def chunk_file(self) -> list[Chunk]:
        """Chunk the given file."""
        raise NotImplementedError


class WholeFileChunker(BaseChunker):
    """Unconditionally chunk the whole file in a single chunk."""

    @staticmethod
    def can_chunk(source_code: str, file_path: Path) -> bool:  # noqa: ARG004
        return True

    def chunk_file(self) -> list[Chunk]:
        return [
            Chunk(
                category=ChunkCategory.whole_file,
                name="<whole_file>",
                content=self.source_code,
                line=1,
            ),
        ]


class PythonChunker(BaseChunker):
    @staticmethod
    def can_chunk(source_code: str, file_path: Path) -> bool:  # noqa: ARG004
        return str(file_path).endswith(".py")

    def chunk_file(self) -> list[Chunk]:
        callables = Callable.from_source_code(self.source_code)
        imports = "\n".join(extract_imports(self.source_code))
        module_level_statements = "\n".join(extract_module_statements(self.source_code))

        chunks: list[Chunk] = []

        if imports.strip() != "":
            chunks.append(
                Chunk(category=ChunkCategory.imports, name="<imports>", line=None, content=imports),
            )
        if module_level_statements.strip() != "":
            chunks.append(
                Chunk(
                    category=ChunkCategory.module_level,
                    name="<module_level_statements>",
                    line=None,
                    content=module_level_statements,
                ),
            )

        chunks.extend(
            Chunk(
                category=ChunkCategory.callable,
                name=callable_.name,
                line=callable_.line,
                content=callable_.code,
            )
            for callable_ in callables
        )
        return chunks


class MarkdownChunker(BaseChunker):
    @staticmethod
    def can_chunk(source_code: str, file_path: Path) -> bool:  # noqa: ARG004
        return str(file_path).endswith(".md")

    def chunk_file(self) -> list[Chunk]:
        chunks: list[Chunk] = []
        current_section: list[str] = []
        current_heading: str | None = None
        current_line = 1
        start_of_section = 1

        for line in self.source_code.split("\n"):
            if line.startswith("#"):
                # If we have a previous section, save it
                if current_section:
                    chunks.append(
                        Chunk(
                            category=ChunkCategory.markdown_section,
                            name=current_heading.replace("#", "").strip()
                            if current_heading is not None
                            else "(no heading)",
                            line=start_of_section,
                            content="\n".join(current_section),
                        ),
                    )
                current_heading = line
                current_section = [line]
                start_of_section = current_line
            else:
                current_section.append(line)
            current_line += 1

        # Add the last section
        if current_section:
            chunks.append(
                Chunk(
                    category=ChunkCategory.markdown_section,
                    name=current_heading.replace("#", "").strip()
                    if current_heading is not None
                    else "(no heading)",
                    line=start_of_section,
                    content="\n".join(current_section),
                ),
            )

        return chunks


class VueChunker(BaseChunker):
    """Chunks Vue Single File Components into their constituent blocks.

    Intention is to put template, script, style (and other custom) blocks
    into their own chunks.
    See https://vuejs.org/api/sfc-spec
    """

    @staticmethod
    def can_chunk(source_code: str, file_path: Path) -> bool:  # noqa: ARG004
        return str(file_path).endswith(".vue")

    def chunk_file(self) -> list[Chunk]:
        # To preserve whitespace, we wrap the source code in a <pre> tag.
        # Without this, BeautifulSoup will strip/fiddle whitespace.
        # See https://stackoverflow.com/a/33788712
        soup_with_pre = BeautifulSoup(
            "<pre>" + self.source_code + "</pre>",
            "html.parser",
        )
        soup_within_pre = soup_with_pre.pre
        chunks: list[Chunk] = []

        # Find all top-level blocks. These are typcially template, script, style,
        # plus any custom blocks.
        top_level_elements = soup_within_pre.find_all(recursive=False)
        for element in top_level_elements:
            chunks.append(
                Chunk(
                    category=ChunkCategory.other,
                    name=element.name,
                    content=str(element),
                    line=element.sourceline,
                ),
            )

        # Get content not in any tag, aggressively stripping whitespace.
        # I'm not sure if it's actually valid to have content outside a <something>
        # tag but whatever doesn't hurt to grab it.
        outer_content_items: list[str] = []
        for outer_content_item in soup_within_pre.find_all(string=True, recursive=False):
            if outer_content_item.strip():
                outer_content_items.append(str(outer_content_item).strip())
        if outer_content_items:
            chunks.append(
                Chunk(
                    category=ChunkCategory.module_level,
                    name="outer_content",
                    content="\n".join(outer_content_items),
                    line=None,
                ),
            )

        return chunks



from collections.abc import Generator
from pathlib import Path

import pytest

from mcpunk import db
from mcpunk.dependencies import deps
from mcpunk.settings import Settings


@pytest.fixture(scope="function", autouse=True)
def fresh_db(tmp_path: Path) -> Generator[None, None, None]:
    """Fiddle settings such that we have a fresh database"""
    settings = Settings(db_path=tmp_path / "test.db")
    assert not settings.db_path.absolute().exists()
    with deps.override(settings_partial=settings):
        db.init_db()
        yield


@pytest.fixture(scope="function", autouse=True)
def fiddle_settings() -> Generator[None, None, None]:
    """Fiddle misc settings for consistency in testing."""
    settings = Settings(
        include_chars_in_response=False,
    )
    with deps.override(settings_partial=settings):
        yield



"""Entry point for running MCPunk.

--------------------------------------------------------------------------------
PRODUCTION
--------------------------------------------------------------------------------

Just
```
{
  "mcpServers": {
    "MCPunk": {
      "command": "/Users/michael/.local/bin/uvx",
      "args": [
        "mcpunk"
      ]
    }
  }
}
```

--------------------------------------------------------------------------------
DEVELOPMENT
--------------------------------------------------------------------------------

Can run on command line with `uvx --from /Users/michael/git/mcpunk --no-cache mcpunk`

Can add to claude like
```
{
  "mcpServers": {
    "MCPunk": {
      "command": "/Users/michael/.local/bin/uvx",
      "args": [
        "--from",
        "/Users/michael/git/mcpunk",
        "--no-cache",
        "mcpunk"
      ]
    }
  }
}
```
"""

# This file is a target for `fastmcp run .../run_mcp_server.py`
import logging
import sys

from mcpunk.db import init_db
from mcpunk.dependencies import Dependencies
from mcpunk.tools import mcp


def _setup_logging() -> logging.Logger:
    """Aggressively control global logging.

    It seems that some MCP clients are messed up by looking at stderr,
    here we remove all log handlers and configure them ourselves with
    careful control.
    """
    settings = Dependencies().settings()

    root_logger = logging.getLogger()

    # Remove any existing handlers
    for handler in root_logger.handlers[:]:
        root_logger.removeHandler(handler)
        handler.close()

    root_logger.setLevel(settings.log_level)

    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s - %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
    )

    if settings.enable_stderr_logging:
        stderr_handler = logging.StreamHandler(sys.stderr)
        stderr_handler.setLevel(settings.log_level)
        stderr_handler.setFormatter(formatter)
        root_logger.addHandler(stderr_handler)

    if settings.enable_log_file:
        log_path = settings.log_file.expanduser().absolute()
        log_path.parent.mkdir(parents=True, exist_ok=True)
        file_handler = logging.FileHandler(log_path)
        file_handler.setLevel(settings.log_level)
        file_handler.setFormatter(formatter)
        root_logger.addHandler(file_handler)

    # Get module logger (will inherit from root)
    return logging.getLogger(__name__)


logger = _setup_logging()
logger.debug("Logging started")


def main() -> None:
    logger.info("Initializing database")
    init_db()
    logger.info("Starting mcp server")
    mcp.run()


if __name__ == "__main__":
    main()



"""Common unset values.

This module exports
1. `Unset`: Singleton object representing an unset state.
2. `AnyUnset`: Alias for `Unset`, typed as `Any` (so essentially ignored by mypy)
3. `UnsetType`: Enum class containing the `Unset` singleton.

This is useful in situations where you need to specify a value as "not defined" or "unset",
but where e.g. `None` is not appropriate (e.g. because `None` is a meaningful value).

Typical uses are like
- `def foo(bar: str | UnsetType = Unset, baz: str | UnsetType):`
  In this we've used `Unset` as we've typed it such that it recognizes `UnsetType`
  as a valid value.
- `def foo(bar: str = AnyUnset):` Here we've used `AnyUnset` since the argument
  is typed as `str`, so passing in `Unset` would raise a type error. But because
  `AnyUnset` is typed as `Any`, it's essentially ignored by mypy.
- API models like `class x(BaseModel): y: int = AnyUnset`
"""

from enum import Enum
from typing import Any

from pydantic_core import core_schema

__all__ = [
    "AnyUnset",
    "Unset",
    "UnsetType",
]


class UnsetType(Enum):
    Unset = object()

    def __repr__(self) -> str:
        return "<Unset>"

    def __str__(self) -> str:
        return self.__repr__()

    def __bool__(self) -> bool:
        return False

    @classmethod
    def __get_pydantic_core_schema__(
        cls,
        _source: Any,
        _handler: Any,
    ) -> core_schema.IsInstanceSchema:
        return core_schema.is_instance_schema(UnsetType)


Unset = UnsetType.Unset
AnyUnset: Any = UnsetType.Unset



# ruff: noqa: E501
from mcpunk.git_analysis import _branches_from_reflog


def test_branches_from_reflog() -> None:
    """Test extracting branches from the reflog contents."""
    branches = _branches_from_reflog(example_reflog, limit=10)
    assert branches == [
        "9a7e727a89db0b0e30eb27c569f8f1ba114b7f59",
        "stratch/7",
        "Avoid-new-try_eval_type-unavailable-on-older-pydantic",
        "decorator_typing",
    ]


def test_branches_from_reflog_limit() -> None:
    """Test branch extraction respects the limit."""
    branches = _branches_from_reflog(example_reflog, limit=3)
    assert branches == [
        "9a7e727a89db0b0e30eb27c569f8f1ba114b7f59",
        "stratch/7",
        "Avoid-new-try_eval_type-unavailable-on-older-pydantic",
    ]


example_reflog = """
9a7e727 (HEAD) HEAD@{0}: checkout: moving from stratch/7 to 9a7e727a89db0b0e30eb27c569f8f1ba114b7f59
97d4034 (stratch/7) HEAD@{1}: checkout: moving from Avoid-new-try_eval_type-unavailable-on-older-pydantic to stratch/7
2c89684 (origin/Avoid-new-try_eval_type-unavailable-on-older-pydantic, Avoid-new-try_eval_type-unavailable-on-older-pydantic) HEAD@{2}: rebase (finish): returning to refs/heads/Avoid-new-try_eval_type-unavailable-on-older-pydantic
2c89684 (origin/Avoid-new-try_eval_type-unavailable-on-older-pydantic, Avoid-new-try_eval_type-unavailable-on-older-pydantic) HEAD@{3}: rebase (start): checkout main
477afc1 HEAD@{4}: cherry-pick: run on all push
2c89684 (origin/Avoid-new-try_eval_type-unavailable-on-older-pydantic, Avoid-new-try_eval_type-unavailable-on-older-pydantic) HEAD@{5}: cherry-pick: Avoid new try_eval_type unavailable on older pydantic
2212d39 (tag: v0.4.0, upstream/main, upstream/HEAD, origin/main, origin/HEAD, main) HEAD@{6}: checkout: moving from stratch/7 to Avoid-new-try_eval_type-unavailable-on-older-pydantic
97d4034 (stratch/7) HEAD@{7}: rebase (finish): returning to refs/heads/stratch/7
97d4034 (stratch/7) HEAD@{8}: rebase (pick): run on all push
e7a99cd HEAD@{9}: rebase (pick): Avoid new try_eval_type unavailable on older pydantic
937b86c HEAD@{10}: rebase (start): checkout fca2db224199adedb80a17841d8bca7540fc1412^
8c058e4 HEAD@{11}: rebase (finish): returning to refs/heads/stratch/7
8c058e4 HEAD@{12}: rebase (pick): Avoid new try_eval_type unavailable on older pydantic
fca2db2 HEAD@{13}: rebase (pick): run on all push
937b86c HEAD@{14}: rebase (pick): Improve decorator typing
2d04b9c HEAD@{15}: rebase (pick): Add py.typed
9a7e727 (HEAD) HEAD@{16}: rebase (squash): Basic Makefile
75032a3 HEAD@{17}: rebase (squash): # This is a combination of 2 commits.
84730d4 HEAD@{18}: rebase (start): checkout 84730d47d148a94b9f3043e9905e446d07ba6db9^
6868c03 HEAD@{19}: rebase (finish): returning to refs/heads/stratch/7
6868c03 HEAD@{20}: rebase (pick): Avoid new try_eval_type unavailable on older pydantic
8013d69 HEAD@{21}: rebase (pick): run on all push
31945d4 HEAD@{22}: rebase (pick): Improve decorator typing
dd59f17 HEAD@{23}: rebase (pick): Add py.typed
4083a3f HEAD@{24}: rebase (pick): Update Makefile
0807c3f (stratch/6) HEAD@{25}: rebase (start): checkout 8051fd44d12d785c84a6b9d1f0403d97021a8816^
c504f8c HEAD@{26}: commit: Avoid new try_eval_type unavailable on older pydantic
07152e2 HEAD@{27}: commit: Update Makefile
9483e5a HEAD@{28}: checkout: moving from decorator_typing to stratch/7
165dd1f (origin/decorator_typing, decorator_typing) HEAD@{29}: rebase (finish): returning to refs/heads/decorator_typing
165dd1f (origin/decorator_typing, decorator_typing) HEAD@{30}: rebase (start): checkout main
8a5e13f HEAD@{31}: cherry-pick: run on all push
165dd1f (origin/decorator_typing, decorator_typing) HEAD@{32}: cherry-pick: Improve decorator typing
8264ed4 HEAD@{33}: cherry-pick: Add py.typed
2212d39 (tag: v0.4.0, upstream/main, upstream/HEAD, origin/main, origin/HEAD, main) HEAD@{34}: checkout: moving from stratch/7 to decorator_typing
9483e5a HEAD@{35}: commit: run on all push
aa25ae5 HEAD@{36}: commit: Improve decorator typing
8051fd4 HEAD@{37}: commit: Add py.typed
0807c3f (stratch/6) HEAD@{38}: checkout: moving from stratch/6 to stratch/7
0807c3f (stratch/6) HEAD@{39}: rebase (finish): returning to refs/heads/stratch/6
0807c3f (stratch/6) HEAD@{40}: rebase (pick): sync-fork Makefile
84730d4 HEAD@{41}: rebase (pick): Basic Makefile
"""



import logging
import os
from pathlib import Path
from string import ascii_lowercase

import pytest

from mcpunk.util import create_file_tree, log_inputs_outputs, matches_filter, rand_str


@pytest.mark.parametrize("log_level", (logging.DEBUG, logging.INFO, "WARNING"))
def test_log_inputs(
    caplog: pytest.LogCaptureFixture,
    log_level: int | str,
) -> None:
    """Test log_inputs decorator captures function inputs and outputs."""
    caplog.set_level(logging.DEBUG)

    @log_inputs_outputs(log_level=log_level)
    def example_func(a: int, *, b: str = "test") -> str:
        return f"{a}{b}"

    result = example_func(1, b="value")
    assert result == "1value"

    expect_log_to_contain_snippets = [
        "Calling tool example_func with inputs:",
        "Arg_0=1",
        "b='value'",
        "resp='1value'",
    ]
    for t in expect_log_to_contain_snippets:
        assert t in caplog.text, t
        for record in caplog.records:
            if t in record.message:
                assert log_level in (record.levelno, record.levelname), record.levelno

    assert "Calling tool example_func with inputs:" in caplog.text
    assert "Arg_0=1" in caplog.text
    assert "b='value'" in caplog.text
    assert "resp='1value'" in caplog.text


def test_create_file_tree(tmp_path: Path) -> None:
    """Test create_file_tree generates correct compact text output."""
    # Set up test files
    (tmp_path / "dir1").mkdir()
    (tmp_path / "dir1/file1.txt").touch()
    (tmp_path / "dir1/other.txt").touch()
    (tmp_path / "dir1/subdir").mkdir()
    (tmp_path / "dir1/subdir/file2.txt").touch()
    (tmp_path / "file3.txt").touch()
    (tmp_path / "dir2/dir4/dir5").mkdir(parents=True)
    (tmp_path / "logs").mkdir()
    (tmp_path / "logs/app.log").touch()
    (tmp_path / "logs/error.log").touch()

    paths = {
        tmp_path / "dir1",
        tmp_path / "dir1/file1.txt",
        tmp_path / "dir1/other.txt",
        tmp_path / "dir1/subdir",
        tmp_path / "dir1/subdir/file2.txt",
        tmp_path / "file3.txt",
        tmp_path / "dir2/dir4/dir5",
        tmp_path / "logs",
        tmp_path / "logs/app.log",
        tmp_path / "logs/error.log",
    }

    result = create_file_tree(
        project_root=tmp_path,
        paths=paths,
    )

    sep = os.path.sep
    expected = (
        f".: file3.txt\ndir1: file1.txt; other.txt\ndir1{sep}subdir: file2.txt\n"
        "logs: app.log; error.log\n"
    )
    assert result == expected


def test_create_file_tree_with_filter(tmp_path: Path) -> None:
    """Test create_file_tree with filter parameter."""
    (tmp_path / "test.py").touch()
    (tmp_path / "main.py").touch()
    (tmp_path / "data.txt").touch()

    paths = {
        tmp_path / "test.py",
        tmp_path / "main.py",
        tmp_path / "data.txt",
    }

    result = create_file_tree(
        project_root=tmp_path,
        paths=paths,
        filter_=[".py"],
    )
    assert result == ".: main.py; test.py\n"


def test_create_file_tree_depth_limit(tmp_path: Path) -> None:
    """Test create_file_tree with depth limit."""
    (tmp_path / "file0.txt").touch()
    (tmp_path / "dir1/dir2/dir3").mkdir(parents=True)
    (tmp_path / "dir1/file1.txt").touch()
    (tmp_path / "dir1/dir2/file2.txt").touch()
    (tmp_path / "dir1/dir2/dir3/file3.txt").touch()

    paths = {
        tmp_path / p
        for p in [
            "file0.txt",
            "dir1",
            "dir1/file1.txt",
            "dir1/dir2",
            "dir1/dir2/file2.txt",
            "dir1/dir2/dir3",
            "dir1/dir2/dir3/file3.txt",
        ]
    }

    result = create_file_tree(
        project_root=tmp_path,
        paths=paths,
        limit_depth_from_root=2,
    )
    assert result == ".: file0.txt\ndir1: file1.txt\n"


def test_rand_str() -> None:
    """Test rand_str generates strings correctly."""
    # Test default length
    result = rand_str()
    assert len(result) == 10
    assert all(c in ascii_lowercase for c in result)

    # Test custom length
    result = rand_str(n=5)
    assert len(result) == 5

    # Test custom characters
    result = rand_str(chars="ABC")
    assert len(result) == 10
    assert all(c in "ABC" for c in result)


def test_matches_filter() -> None:
    """Test matches_filter with different filter types."""
    # Test None filter
    assert matches_filter(None, "test") is True
    assert matches_filter(None, None) is True

    # Test list filter
    assert matches_filter(["abc", "def"], "abcdef") is True
    assert matches_filter(["xyz", "123"], "abcdef") is False
    assert matches_filter(["abc"], None) is False



import enum
import logging
from hashlib import sha256
from typing import Literal, assert_never, get_args

from pydantic import (
    BaseModel,
    Field,
)

from mcpunk.util import matches_filter

logger = logging.getLogger(__name__)


class ChunkCategory(enum.StrEnum):
    callable = "callable"
    markdown_section = "markdown section"
    imports = "imports"
    module_level = "module_level"
    whole_file = "whole_file"
    other = "other"


# Seems if you annotate a FastMCP tool function with an enum it totally
# crashes claude desktop. So define an equivalent Literal type here.
ChunkCategoryLiteral = Literal[
    "callable",
    "markdown section",
    "imports",
    "module_level",
    "whole_file",
    "other",
]
assert set(get_args(ChunkCategoryLiteral)) == set(ChunkCategory.__members__.values())


class Chunk(BaseModel):
    """A chunk of a file, e.g. a function or a markdown section."""

    category: ChunkCategory = Field(description="`function`, `markdown section`, `imports`")
    name: str = Field(description="`my_function` or `MyClass` or `# My Section`")
    line: int | None = Field(description="Line within file where it starts. First line is 1.")
    content: str = Field(description="Content of the chunk")

    @property  # Consider caching
    def id_(self) -> str:
        """Generate a (probably) unique ID based on content, name, line, and category.

        This approach means that the ID will stay the same even if the chunk is recreated
        (e.g. as opposed to a totally random ID).
        There's chance of ids being duplicated, especially across files.
        The id includes the chunk name, which makes debugging and monitoring of tool
        requests far nicer.
        """
        components = [self.content, self.name, str(self.line), str(self.category)]
        return self.name + "_" + sha256("".join(components).encode()).hexdigest()[:10]

    def matches_filter(
        self,
        filter_: None | list[str] | str,
        filter_on: Literal["name", "content", "name_or_content"],
    ) -> bool:
        """Return True if the chunk's name matches the given filter.

        str matches if the chunk's name contains the string.
        list[str] matches if the chunk's name contains any of the strings in the list.
        None matches all chunks.
        """
        if filter_on == "name":
            data = self.name
        elif filter_on == "content":
            data = self.content
        elif filter_on == "name_or_content":
            data = self.content + self.name
        else:
            assert_never(filter_on)
        return matches_filter(filter_, data)



from pathlib import Path

from mcpunk.file_chunkers import VueChunker


def test_vue_chunker() -> None:
    source_code = """\
<template>
  <div class="example">{{ msg }}</div>
</template>

<script lang="ts">
export default {
  data() {
    return {
      msg: 'Hello world!'
    }
  }
}
</script>

blah

<style>
.example {
  color: red;
}
</style>

<custom1>
  This could be e.g. documentation for the component.
</custom1>
something_else
    """
    assert VueChunker.can_chunk("", Path("anything.vue"))
    assert not VueChunker.can_chunk("", Path("blah.vues"))

    chunks = VueChunker(source_code, Path("anything.vue")).chunk_file()
    chunks = sorted(
        chunks,
        key=lambda x: x.line if x.line is not None else 999999,
    )

    assert [x.name for x in chunks] == [
        "template",
        "script",
        "style",
        "custom1",
        "outer_content",
    ]

    # Test line numbers
    assert chunks[0].line == 1  # template
    assert chunks[1].line == 5  # script
    assert chunks[2].line == 17  # style
    assert chunks[3].line == 23  # custom1

    # Test content
    assert (
        chunks[0].content
        == """\
<template>
  <div class="example">{{ msg }}</div>
</template>"""
    )

    assert (
        chunks[1].content
        == """\
<script lang="ts">
export default {
  data() {
    return {
      msg: 'Hello world!'
    }
  }
}
</script>"""
    )

    assert (
        chunks[2].content
        == """\
<style>
.example {
  color: red;
}
</style>"""
    )

    assert (
        chunks[3].content
        == """\
<custom1>
  This could be e.g. documentation for the component.
</custom1>"""
    )

    # Test outer content combines non-tag content
    assert chunks[4].content.strip() == "blah\nsomething_else"

    # Test categories
    assert all(chunk.category == "other" for chunk in chunks[:4])
    assert chunks[4].category == "module_level"



from pathlib import Path

import pytest

from mcpunk.file_chunk import ChunkCategory
from mcpunk.file_chunkers import PythonChunker


def test_python_chunker_basic() -> None:
    """A basic test that it picks out key elements.

    Notably the picking apart of python code into callables, imports, and module-level
    statements is tested elsewhere.
    """
    source_code = """\
from typing import List
import os

x = 1

def func1(a: int) -> str:
    return str(a)

y = 2

class MyClass:
    def method1(self) -> None:
        pass

    @property
    def prop1(self) -> int:
        return 42
"""
    assert PythonChunker.can_chunk("", Path("test.py"))
    assert not PythonChunker.can_chunk("", Path("test.txt"))

    chunks = PythonChunker(source_code, Path("test.py")).chunk_file()
    chunks = sorted(chunks, key=lambda x: x.line if x.line is not None else -1)

    assert [x.name for x in chunks] == [
        "<imports>",
        "<module_level_statements>",
        "func1",
        "MyClass",
        "method1",
        "prop1",
    ]

    # Test categories
    assert chunks[0].category.value == "imports"
    assert chunks[1].category.value == "module_level"
    assert chunks[2].category.value == "callable"
    assert chunks[3].category.value == "callable"
    assert chunks[4].category.value == "callable"
    assert chunks[5].category.value == "callable"

    # Test content
    assert chunks[0].content == "from typing import List\nimport os"
    assert chunks[1].content == "x = 1\ndef func1...\ny = 2\nclass MyClass..."

    assert chunks[2].content == "def func1(a: int) -> str:\n    return str(a)"
    assert chunks[2].line == 6

    assert "class MyClass:" in chunks[3].content
    assert chunks[3].line == 11

    assert "def method1(self)" in chunks[4].content
    assert chunks[4].line == 12

    assert "@property" in chunks[5].content
    assert "def prop1(self)" in chunks[5].content
    assert chunks[5].line == 16


def test_python_chunker_empty() -> None:
    """Tests empty source files produce empty chunk lists."""
    empty_source = ""
    chunks = PythonChunker(empty_source, Path("test.py")).chunk_file()
    assert chunks == []


def test_python_chunker_only_imports() -> None:
    """Tests files with only imports."""
    imports_only = """\
from typing import List
import os
"""
    chunks = PythonChunker(imports_only, Path("test.py")).chunk_file()
    assert len(chunks) == 1
    assert chunks[0].name == "<imports>"
    assert chunks[0].content == "from typing import List\nimport os"


def test_python_chunker_only_module_level() -> None:
    """Tests files with only module level statements."""
    module_level_only = """\
x = 1
y = 2
"""
    chunks = PythonChunker(module_level_only, Path("test.py")).chunk_file()
    assert len(chunks) == 1
    assert chunks[0].name == "<module_level_statements>"
    assert chunks[0].content == "x = 1\ny = 2"


def test_python_chunker_only_callables() -> None:
    """Tests files with only function/class definitions."""
    callables_only = """\
def func1() -> None:
    pass

class MyClass:
    def method1(self) -> None:
        pass
"""
    chunks = PythonChunker(callables_only, Path("test.py")).chunk_file()
    assert len(chunks) == 4
    assert [x.name for x in chunks] == ["<module_level_statements>", "func1", "MyClass", "method1"]
    module_level = chunks[0]
    assert module_level.category == ChunkCategory.module_level
    # We should still have module level statements, just with skeleton of the callables:
    assert module_level.content == "def func1...\nclass MyClass..."


def test_python_chunker_invalid_syntax() -> None:
    with pytest.raises(Exception, match=".*"):
        _ = PythonChunker("x = (1", Path("test.py")).chunk_file()



from pathlib import Path

from mcpunk.file_chunkers import MarkdownChunker


def test_markdown_chunker() -> None:
    source_code = """\
Pre-header content

# Section 1
Some content

## Subsection 1.1
More content
More lines

# Section 2
Final content
"""
    assert MarkdownChunker.can_chunk("", Path("test.md"))
    assert not MarkdownChunker.can_chunk("", Path("test.txt"))

    chunks = MarkdownChunker(source_code, Path("test.md")).chunk_file()
    chunks = sorted(chunks, key=lambda x: x.line if x.line is not None else -1)

    assert [x.name for x in chunks] == ["(no heading)", "Section 1", "Subsection 1.1", "Section 2"]

    # Test categories
    assert all(x.category.value == "markdown section" for x in chunks)

    # Test line numbers
    assert chunks[0].line == 1  # No heading section
    assert chunks[1].line == 3  # Section 1
    assert chunks[2].line == 6  # Subsection 1.1
    assert chunks[3].line == 10  # Section 2

    # Test content
    assert chunks[0].content == "Pre-header content\n"
    assert chunks[1].content == "# Section 1\nSome content\n"
    assert chunks[2].content == "## Subsection 1.1\nMore content\nMore lines\n"
    assert chunks[3].content == "# Section 2\nFinal content\n"



import enum
from collections.abc import Generator
from contextlib import contextmanager
from datetime import UTC, datetime
from typing import Any

import sqlalchemy as sa
from sqlalchemy import DateTime, Enum, String, TypeDecorator
from sqlalchemy.orm import (
    DeclarativeBase,
    Mapped,
    Session,
    mapped_column,
)

from mcpunk.dependencies import deps
from mcpunk.unset import AnyUnset

CURRENT_DB_VERSION = "1"


def _remove_unset(d: dict[str, Any]) -> dict[str, Any]:
    """Return new dict with any key where the val is `shared_utils.unset.Unset` removed"""
    return {k: v for k, v in d.items() if v is not AnyUnset}


class TZDateTime(TypeDecorator):  # type: ignore[type-arg]
    impl = DateTime
    cache_ok = True

    def process_bind_param(self, value: datetime | None, dialect: Any) -> datetime | None:
        _ = dialect
        if value is not None and value.tzinfo is None:
            raise ValueError("DateTime must be timezone-aware")
        return value


class Base(DeclarativeBase):
    created: Mapped[datetime] = mapped_column(TZDateTime(), default=lambda: datetime.now(UTC))


@enum.unique
class TaskState(enum.StrEnum):
    TODO = "todo"
    DOING = "doing"
    DONE = "done"


@enum.unique
class TaskFollowUpCriticality(enum.StrEnum):
    NO_FOLLOWUP = "no_followup"
    LOW = "low"
    MEDIUM = "medium"
    HIGH = "high"


class Task(Base):
    __tablename__ = "task"

    id: Mapped[int] = mapped_column(primary_key=True)
    state: Mapped[TaskState] = mapped_column(
        Enum(
            TaskState,
            native_enum=False,
            create_constraint=True,
            values_callable=lambda py_enum: [e.value for e in py_enum],
        ),
        default=TaskState.TODO,
    )
    last_picked_up_at: Mapped[datetime | None] = mapped_column(TZDateTime(), nullable=True)
    action: Mapped[str] = mapped_column(String(2_000))
    outcome: Mapped[str | None] = mapped_column(String(10_000), nullable=True)
    follow_up_criticality: Mapped[TaskFollowUpCriticality] = mapped_column(
        Enum(
            TaskFollowUpCriticality,
            native_enum=False,
            create_constraint=True,
            values_callable=lambda py_enum: [e.value for e in py_enum],
        ),
        nullable=True,
    )

    @staticmethod
    def create(
        *,
        action: str,
    ) -> "Task":
        kwargs: dict[str, Any] = {
            "action": action,
        }
        return Task(**_remove_unset(kwargs))


class DBVersion(Base):
    __tablename__ = "db_version"
    id: Mapped[int] = mapped_column(primary_key=True)
    version = mapped_column(String(32))


class TaskManager:
    def __init__(self, session: Session) -> None:
        self.sess = session

    def add_task(self, action: str) -> None:
        task = Task.create(action=action)
        self.sess.add(task)

    def get_task(self) -> Task | None:
        stmt = (
            sa.select(Task)
            .where(
                sa.or_(
                    Task.state == TaskState.TODO,
                    sa.and_(
                        Task.state == TaskState.DOING,
                        sa.or_(
                            Task.last_picked_up_at == None,  # noqa: E711
                            Task.last_picked_up_at
                            < datetime.now(UTC) - deps.settings().task_queue_visibility_timeout,
                        ),
                    ),
                ),
            )
            .order_by(Task.created)
            .limit(1)
        )
        task = self.sess.execute(stmt).scalar_one_or_none()
        if task:
            task.state = TaskState.DOING
            task.last_picked_up_at = datetime.now(UTC)
        return task

    def set_task_done(
        self,
        task_id: int,
        outcome: str,
        follow_up_criticality: TaskFollowUpCriticality,
    ) -> None:
        task = self.sess.get(Task, task_id)
        if not task:
            raise ValueError(f"No task found with id {task_id}")

        task.state = TaskState.DONE
        task.outcome = outcome
        task.follow_up_criticality = follow_up_criticality


@contextmanager
def get_task_manager() -> Generator[TaskManager, None, None]:
    make_session = deps.session_maker()
    with make_session.begin() as sess:
        yield TaskManager(sess)


def _init_pragmas() -> None:
    """Call me liberally why not."""
    with deps.session_maker().begin() as sess:
        for stmt in [
            "PRAGMA journal_mode = WAL;",
            "PRAGMA synchronous = NORMAL;",
            "PRAGMA busy_timeout = 5000;",
            "PRAGMA cache_size = -20000;",
            "PRAGMA foreign_keys = ON;",
        ]:
            sess.execute(sa.text(stmt))


def _init_totally_new_db() -> None:
    db_path = deps.settings().db_path
    assert not db_path.exists()
    db_path.parent.mkdir(parents=True, exist_ok=True)
    _init_pragmas()
    with deps.session_maker().begin() as sess:
        Base.metadata.create_all(sess.get_bind())
        sess.add(DBVersion(version=CURRENT_DB_VERSION))


def init_db() -> None:
    if not deps.settings().db_path.exists():
        _init_totally_new_db()
    _init_pragmas()
    with deps.session_maker().begin() as sess:
        version = sess.scalars(sa.select(DBVersion.version)).one_or_none()
    if version != CURRENT_DB_VERSION:
        raise ValueError(
            f"Database version mismatch. Expected {CURRENT_DB_VERSION}, "
            f"got {version}. You might like to just delete the database and "
            f"it'll be automatically recreated.",
        )



# ruff: noqa: E402
from collections.abc import Generator
from contextlib import contextmanager
from dataclasses import dataclass
from typing import Any, Union

from sqlalchemy import Engine, create_engine
from sqlalchemy.orm import (
    Session,
    sessionmaker,
)

__all__ = [
    "Dependencies",
    "deps",
]


class Singleton(type):
    _instances: dict[Any, Any] = {}  # noqa: RUF012

    def __call__(cls, *args: Any, **kwargs: Any) -> Any:
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


@dataclass
class DependencyState:
    settings: Union["Settings", None] = None
    engine: Engine | None = None
    session_maker: sessionmaker[Session] | None = None

    settings_override: Union["Settings", None] = None
    engine_override: Engine | None = None
    session_maker_override: sessionmaker[Session] | None = None


class Dependencies(metaclass=Singleton):
    """Dependencies that can play nice with testing.

    If you want to get settings, db session, etc. do it here. This means that in
    tests we can patch/etc JUST this object and have everything play nice.
    """

    def __init__(self) -> None:
        self._state = DependencyState()

    def settings(self) -> "Settings":
        if self._state.settings_override is not None:
            return self._state.settings_override
        if self._state.settings is None:
            self._state.settings = Settings()
        return self._state.settings

    def db_engine(self) -> "Engine":
        if self._state.engine_override is not None:
            return self._state.engine_override
        if self._state.engine is None:
            settings = self.settings()
            self._state.engine = create_engine(
                f"sqlite:///{settings.db_path}?check_same_thread=true&timeout=10&uri=true",
            )
        return self._state.engine

    def session_maker(self) -> sessionmaker[Session]:
        if self._state.session_maker_override is not None:
            return self._state.session_maker_override
        if self._state.session_maker is None:
            self._state.session_maker = sessionmaker(
                autocommit=False,
                expire_on_commit=True,
                autoflush=False,
                bind=self.db_engine(),
            )
        return self._state.session_maker

    @contextmanager
    def override(
        self,
        settings: Union["Settings", None] = None,
        db_engine: Engine | None = None,
        session_maker: sessionmaker[Session] | None = None,
        settings_partial: Union["Settings", None] = None,
    ) -> Generator[None, None, None]:
        """Override dependency functions for testing."""
        # Backup current state and methods
        orig_state = self._state

        if settings_partial is not None:
            if settings is not None:
                # heyo maybe an @override 🤷
                raise ValueError("settings and settings_partial cannot both be set")

            if orig_state.settings_override is not None:
                existing_kwargs = orig_state.settings_override.model_dump()
            else:
                existing_kwargs = {}
            new_kwargs = settings_partial.model_dump(exclude_unset=True)
            merged_kwargs = {**existing_kwargs, **new_kwargs}
            settings = Settings(**merged_kwargs)

        # Carry across the overrides, typical use for this is multiple nesting
        # of override contexts managers.
        new_state = DependencyState(
            settings_override=orig_state.settings_override,
            engine_override=orig_state.engine_override,
            session_maker_override=orig_state.session_maker_override,
        )
        new_state.settings_override = settings
        new_state.engine_override = db_engine
        new_state.session_maker_override = session_maker
        self._state = new_state

        try:
            yield
        finally:
            self._state = orig_state


from mcpunk.settings import Settings

deps = Dependencies()



from pathlib import Path

from git import Repo


def get_recent_branches(repo_path: Path, limit: int = 10) -> list[str]:
    """Get the list of up to `limit` recently checked out branches in the given git repo."""
    repo = Repo(repo_path)
    reflog = repo.git.reflog()
    branches = _branches_from_reflog(reflog, limit)
    return branches


def _branches_from_reflog(reflog: str, limit: int) -> list[str]:
    branches = []
    for line in reflog.splitlines():
        if "checkout:" in line:
            branch = line.split("to ")[-1].strip()
            if branch not in branches:
                branches.append(branch)
                if len(branches) >= limit:
                    break
    return branches



import logging
import random
from collections import defaultdict
from collections.abc import Callable
from copy import deepcopy
from functools import wraps
from pathlib import Path
from string import ascii_lowercase
from typing import ParamSpec, TypeVar, assert_never

logger = logging.getLogger(__name__)

P = ParamSpec("P")
R = TypeVar("R")


def log_inputs_outputs(
    log_level: int | str = logging.INFO,
) -> Callable[[Callable[P, R]], Callable[P, R]]:
    """Decorator to wrap a tool function and log its inputs and outputs.

    mcp = FastMCP()
    @mcp.tool()
    @log_inputs()
    def get_a_joke(): ...

    Args:
        log_level: The log level to use for the log messages, like `logging.INFO`
            or "INFO", matching those in the `logging` module.
    """
    if isinstance(log_level, str):
        level = logging.getLevelNamesMapping()[log_level]
    else:
        level = log_level

    def decorator(func: Callable[P, R]) -> Callable[P, R]:
        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            lines = [
                "",
                " " * 2 + "./" + "-" * 116 + "\\.",
                " " * 1 + "./" + " " * 118 + "\\.",
                " " * 0 + "./" + " " * 120 + "\\.",
                f"Calling tool {func.__name__} with inputs:",
            ]
            for i, v in enumerate(args):
                lines.append(f"    Arg_{i}={v!r}")
            for k, v in kwargs.items():
                lines.append(f"    {k}={v!r}")
            logger.log(level, "\n".join(lines))
            resp = func(*args, **kwargs)
            lines = [
                "",
                f"    resp={resp!r}",
                " " * 0 + ".\\" + " " * 120 + "/.",
                " " * 1 + ".\\" + " " * 118 + "/.",
                " " * 2 + ".\\" + "-" * 116 + "/.",
            ]
            logger.log(level, "\n".join(lines))
            return resp

        return wrapper

    return decorator


def create_file_tree(
    *,
    project_root: Path,
    paths: set[Path],
    limit_depth_from_root: int | None = None,
    filter_: None | list[str] | str = None,
) -> str | None:
    """Create a compact text representation of files in a directory structure.

    Outputs files grouped by their parent directory, with each directory and its
    files on a single line. Only includes files, not directories.
    e.g.
    ```
    dir1: file1.txt
    dir2/dir3: file2.txt; file3.txt
    ...
    ```

    Args:
        project_root: The root directory of the project.
        paths: Set of paths to potentially include in the output.
        limit_depth_from_root: If provided, exclude files deeper than this many
            levels from the root directory.
        filter_: If provided, only include paths containing any of these strings.
            None matches all paths.

    Returns:
        None if no files match the criteria, otherwise a string where each line is:
        "{relative_directory_path}: file1.txt; file2.txt; file3.txt"
    """
    paths = deepcopy(paths)  # Avoid mutation shenanigans
    project_root = project_root.absolute()

    filtered_paths = {
        x
        for x in paths
        if matches_filter(filter_, str(x))
        and x.is_file()
        and (
            limit_depth_from_root is None
            or _get_depth_from_root(project_root, x) <= limit_depth_from_root
        )
    }

    if len(filtered_paths) == 0:
        return None

    response = ""
    files_by_parent_dir: dict[Path, list[Path]] = defaultdict(list)
    for p in sorted(filtered_paths):
        files_by_parent_dir[p.parent].append(p)
    for parent_dir in sorted(files_by_parent_dir.keys()):
        response += f"{parent_dir.relative_to(project_root)}: "
        response += "; ".join(x.name for x in files_by_parent_dir[parent_dir])
        response += "\n"
    return response


def _get_depth_from_root(root: Path, file: Path) -> int:
    return len(file.relative_to(root).parts)


def rand_str(n: int = 10, chars: str = ascii_lowercase) -> str:
    return "".join(random.choice(chars) for _ in range(n))


def matches_filter(filter_: None | list[str] | str, data: str | None) -> bool:
    """Return True if the data matches the given filter.

    filter_ can be:
    - None matches all data
    - str matches if the data contains the string
    - list[str] matches if the data contains any of the strings in the list

    if data is None it never matches (unless filter_ is None)
    """
    if filter_ is None:
        return True
    if data is None:
        return False
    if isinstance(filter_, str):
        return filter_ in data
    if isinstance(filter_, list):
        return any(x in data for x in filter_)
    assert_never(filter_)



import inspect
import json
import logging
import pathlib
import textwrap
from collections.abc import Sequence
from typing import Annotated, Any, Literal, Self, assert_never

import mcp.types as mcp_types
from fastmcp import FastMCP
from git import Repo
from pydantic import (
    BaseModel,
    Field,
    model_validator,
)
from pydantic_core import to_jsonable_python

from mcpunk import db
from mcpunk.dependencies import deps
from mcpunk.file_breakdown import (
    File,
)
from mcpunk.file_breakdown import (
    Project as FileBreakdownProject,
)
from mcpunk.git_analysis import get_recent_branches
from mcpunk.util import create_file_tree, log_inputs_outputs

logger = logging.getLogger(__name__)

PROJECTS: dict[str, "ToolProject"] = {}


mcp = FastMCP("Code Analysis")

ToolResponseSingleItem = mcp_types.TextContent | mcp_types.ImageContent | mcp_types.EmbeddedResource
ToolResponseSequence = Sequence[ToolResponseSingleItem]
ToolResponse = ToolResponseSequence | ToolResponseSingleItem
FilterType = Annotated[
    str | list[str] | None,
    Field(description="Match if any of these strings appear. Match all if None/null."),
]


class ToolProject(BaseModel):
    """A project containing files split into chunks and so on.

    These are created by the `configure_project` tool, and can be referenced by name
    (which is the key in the `PROJECTS` global dict) when calling other tools.
    """

    chunk_project: FileBreakdownProject

    @property
    def root(self) -> pathlib.Path:
        return self.chunk_project.root

    @property
    def git_path(self) -> pathlib.Path:
        if str(self.root).endswith(".git"):
            git_dir_path = self.root
        else:
            git_dir_path = self.root / ".git"
        if not git_dir_path.exists():
            raise ValueError(f"git dir not found at {git_dir_path}")
        return git_dir_path


class ProjectFile(BaseModel):
    project_name: str
    rel_path: Annotated[pathlib.Path, Field(description="Relative to project root")]

    @property
    def project(self) -> ToolProject:
        return _get_project_or_error(self.project_name)

    @property
    def abs_path(self) -> pathlib.Path:
        return self.project.chunk_project.root / self.rel_path

    @property
    def file(self) -> File:
        abs_path = self.abs_path
        matching_files = [f for f in self.project.chunk_project.files if f.abs_path == abs_path]
        if len(matching_files) != 1:
            raise ValueError(f"File {self.abs_path} not found in project {self.project_name}")
        return matching_files[0]

    @model_validator(mode="after")
    def validate_misc(self) -> Self:
        assert self.project is not None
        assert self.file is not None
        return self


class MCPToolOutput(BaseModel):
    """The output of a tool.

    You can specify any of the items in here, and they will all be rendered and
    returned to the client. If you specify NOTHING then the default response
    will be returned.
    """

    is_error: bool = False
    # Anything that pydantic core to_jsonable_python can handle - that's a lot of stuff!
    jsonable: Any | None = None
    raw: ToolResponse | None = None
    text: str | None = None

    # You might like this set to 2/4 for debugging, makes things look nice!
    # But that means more token usage I guess.
    # an int will do what you expect. None for compact. If unset, will default to
    # the value from settings.
    indent: int | Literal["no_indent"] = Field(
        default_factory=lambda: deps.settings().default_response_indent,
    )

    default_response: str = "No response provided. This is not an error."

    # If the sum of the length of all text responses is greater than this
    # then an error will be returned to the caller. non-text responses (image, etc)
    # are not counted.
    max_chars: int = Field(
        default_factory=lambda: deps.settings().default_response_max_chars,
    )

    # Whether to include the number of characters in the response in the response.
    # So like `[DEBUG INFO: Response is 1234 chars]` prefixed to the response.
    include_chars_in_response: bool = Field(
        default_factory=lambda: deps.settings().include_chars_in_response,
    )

    def render(self) -> ToolResponse:
        indent: int | None
        if self.indent == "no_indent":
            indent = None
        elif isinstance(self.indent, int):
            assert isinstance(self.indent, int)
            indent = self.indent
        else:
            assert_never(self.indent)
        assert indent is None or isinstance(indent, int)

        out: list[ToolResponseSingleItem] = []
        if self.is_error:
            out.append(mcp_types.TextContent(type="text", text="An error occurred."))
        if self.jsonable is not None:
            logger.debug(
                "Jsonable response\n"
                + textwrap.indent(json.dumps(to_jsonable_python(self.jsonable), indent=2), "    "),
            )
            out.append(
                mcp_types.TextContent(
                    type="text",
                    text=json.dumps(to_jsonable_python(self.jsonable), indent=indent),
                ),
            )
        if self.raw is not None:
            if isinstance(self.raw, ToolResponseSingleItem):
                out.append(self.raw)
            elif isinstance(self.raw, Sequence):
                assert all(isinstance(x, ToolResponseSingleItem) for x in self.raw)
                out.extend(self.raw)
            else:
                assert_never(self.raw)
        if self.text is not None:
            out.append(mcp_types.TextContent(type="text", text=self.text))
        if len(out) == 0:
            # Use default response if no data was provided
            assert not self.is_error  # Don't want to say there's an error if there was!
            out.append(mcp_types.TextContent(type="text", text=self.default_response))

        total_chars = sum(len(x.text) for x in out if isinstance(x, mcp_types.TextContent))
        if total_chars > self.max_chars:
            msg = (
                f"Response is {total_chars} chars which exceed the maximum allowed "
                f"of {self.max_chars}. Please adjust your request and try again."
            )
            logger.warning(msg)
            out = [mcp_types.TextContent(type="text", text=msg)]

        if self.include_chars_in_response:
            len_msg = f"[DEBUG INFO: Response is {total_chars} chars]"
            if len(out) == 1 and isinstance(out[0], mcp_types.TextContent):
                out[0].text = f"{len_msg}\n\n{out[0].text}"
            else:
                out.insert(
                    0,
                    mcp_types.TextContent(type="text", text=f"Response is {total_chars} chars"),
                )

        final_out: ToolResponse
        if len(out) == 1:
            final_out = out[0]
        else:
            final_out = out
        # logger.debug(f"Response {final_out}")
        logger.debug(
            "Final response\n"
            + textwrap.indent(json.dumps(to_jsonable_python(final_out), indent=2), "    "),
        )
        return final_out


@mcp.tool()
@log_inputs_outputs()
def get_a_joke(animal: Annotated[str, Field(max_length=20)]) -> ToolResponse:
    """Get a really funny joke! For testing :)"""
    return MCPToolOutput(
        text=(
            f"Why did the {animal} cross the road?\n"
            f"To get to the other side!\n"
            f"Because it was a {animal}."
        ),
    ).render()


@mcp.tool()
@log_inputs_outputs()
def configure_project(
    root_path: Annotated[pathlib.Path, Field(description="Root path of the project")],
    project_name: Annotated[
        str,
        Field(
            description=(
                "Name of the project, for you to pick buddy, "
                "something short and sweet and memorable and unique"
            ),
        ),
    ],
) -> ToolResponse:
    """Configure a new project containing files.

    Each file in the project is split into 'chunks' - logical sections like functions,
    classes, markdown sections, and import blocks.

    After configuring, a common workflow is:
    1. list_all_files_in_project to get an overview of the project (with
       an initial limit on the depth of the search)
    2. Find files by function/class definition:
       find_files_by_chunk_content(... ["def my_funk"])
    3. Find files by function/class usage:
       find_files_by_chunk_content(... ["my_funk"])
    4. Determine which chunks in the found files are relevant:
        find_matching_chunks_in_file(...)
    5. Get details about the chunks:
       chunk_details(...)

    Use ~ (tilde) literally if the user specifies it in paths.
    """
    path = root_path.expanduser().absolute()
    if project_name in PROJECTS:
        raise ValueError(f"Project {project_name} already exists")
    project = ToolProject(chunk_project=FileBreakdownProject.from_root_dir(path))
    PROJECTS[project_name] = project
    return MCPToolOutput(
        text=(
            inspect.cleandoc(f"""\
            Project {path} configured with {len(project.chunk_project.files)} files.
            Files are split into 'chunks' - logical sections like:
            - Functions (e.g. 'def my_function')
            - Classes (e.g. 'class MyClass')
            - Markdown sections (e.g. '# Section')
            - Import blocks

            After configuring, a common workflow is:
            1. list_all_files_in_project to get an overview of the project (with
               an initial limit on the depth of the search)
            2. Find files by function/class definition:
               find_files_by_chunk_content(... ["def my_funk"])
            3. Find files by function/class usage:
               find_files_by_chunk_content(... ["my_funk"])
            4. Determine which chunks in the found files are relevant:
                find_matching_chunks_in_file(...)
            5. Get details about the chunks:
               chunk_details(...)

            Do not immediately list files or otherwise use the project
            unless explicitly told to do so.
        """)
        ),
    ).render()


@mcp.tool()
@log_inputs_outputs()
def list_all_files_in_project(
    project_name: str,
    path_filter: FilterType = None,
    limit_depth_from_root: Annotated[
        int | None,
        Field(
            description=(
                "Limit the depth of the search to this many directories from the root. "
                "Typically,start with 1 to get an overview of the project."
                "If None, search all directories from the root."
            ),
        ),
    ] = None,
) -> ToolResponse:
    """List all files in a project, returning a file tree.

    This is useful for getting an overview of the project, or specific
    subdirectories of the project.

    A project may have many files, so you are suggested
    to start with a depth limit to get an overview, and then continue increasing
    the depth limit with a filter to look at specific subdirectories.
    """
    project = _get_project_or_error(project_name)
    data = create_file_tree(
        project_root=project.root,
        paths={x.abs_path for x in project.chunk_project.files},
        limit_depth_from_root=limit_depth_from_root,
        filter_=path_filter,
    )
    if data is None:
        return MCPToolOutput(text="No paths").render()
    elif isinstance(data, str):
        return MCPToolOutput(text=data).render()
    else:
        assert_never(data)


@mcp.tool()
@log_inputs_outputs()
def find_files_by_chunk_content(
    project_name: str,
    chunk_contents_filter: FilterType,
) -> ToolResponse:
    """Step 1: Find files containing chunks with matching text.

    Returns file tree only showing which files contain matches.
    You must use find_matching_chunks_in_file on each relevant file
    to see the actual matches.

    Example workflow:
    1. Find files:
       files = find_files_by_chunk_content(project, ["MyClass"])
    2. For each file, find actual matches:
       matches = find_matching_chunks_in_file(file, ["MyClass"])
    3. Get content:
       content = chunk_details(file, match_id)
    """
    return _filter_files_by_chunk(project_name, chunk_contents_filter, "name_or_content").render()


@mcp.tool()
@log_inputs_outputs()
def find_matching_chunks_in_file(
    project_name: str,
    rel_path: Annotated[pathlib.Path, Field(description="Relative to project root")],
    filter_: FilterType,
) -> ToolResponse:
    """Step 2: Find the actual matching chunks in a specific file.

    Required after find_files_by_chunk_content or list_all_files_in_project to see
    matches, as those tools only show files, not their contents.

    This can be used for things like:
      - Finding all chunks in a file that make reference to a specific function
        (e.g. find_matching_chunks_in_file(..., ["my_funk"])
      - Finding a chunk where a specific function is defined
        (e.g. find_matching_chunks_in_file(..., ["def my_funk"])

    Returns array of {n: name, t: type, id: identifier, chars: length}
    """
    proj_file = ProjectFile(project_name=project_name, rel_path=rel_path)
    return _list_chunks_in_file(proj_file, filter_, "name_or_content").render()


@mcp.tool()
@log_inputs_outputs()
def chunk_details(
    project_name: str,
    rel_path: Annotated[pathlib.Path, Field(description="Relative to project root")],
    chunk_id: str,
) -> ToolResponse:
    """Get full content of a specific chunk.

    Returns chunk content as string.

    Common patterns:
    1. Final step after find_matching_chunks_in_file finds relevant chunks
    2. Examining implementations after finding definitions/uses
    """
    proj_file = ProjectFile(project_name=project_name, rel_path=rel_path)
    target_file = proj_file.file
    chunks_raw = [chunk for chunk in target_file.chunks if chunk.id_ == chunk_id]
    chunk_contents = [inspect.cleandoc(x.content) for x in chunks_raw]
    if len(chunk_contents) == 0:
        return MCPToolOutput(
            text="No matching chunks. Please use other tools to find available chunks.",
        ).render()
    elif len(chunk_contents) == 1:
        return MCPToolOutput(text=chunk_contents[0]).render()
    else:
        # Honestly should perhaps just raise here 🤷
        resp = "WARNING MULTIPLE CHUNKS FOUND THIS IS VERY ODD"
        for i, chunk_content in enumerate(chunk_contents):
            resp += f"\n\n# Chunk {i + 1}{chunk_content}"
        return MCPToolOutput(text=resp).render()


@mcp.tool()
@log_inputs_outputs()
def list_most_recently_checked_out_branches(
    project_name: str,
    n: Annotated[int, Field(ge=20, le=50)] = 20,
) -> ToolResponse:
    """List the n most recently checked out branches in the project"""
    project = _get_project_or_error(project_name)
    return MCPToolOutput(text="\n".join(get_recent_branches(project.git_path, n))).render()


@mcp.tool()
@log_inputs_outputs()
def diff_with_ref(
    project_name: str,
    ref: Annotated[str, Field(max_length=100)],
) -> ToolResponse:
    """Return a summary of the diff between HEAD and the given ref.

    You probably want the ref  to be the 'base' branch like develop or main, off which
    PRs are made - and you can likely determine this by viewing the most recently
    checked out branches.
    """
    project = _get_project_or_error(project_name)
    repo = Repo(project.git_path)
    # head = repo.head.commit
    # compare_from = repo.commit(ref)
    # diffs = compare_from.diff(head, create_patch=True)
    # print(repo.git.diff(f"{ref}s...HEAD", ignore_blank_lines=True, ignore_space_at_eol=True))
    diff = repo.git.diff(
        f"{ref}...HEAD",
        ignore_blank_lines=True,
        ignore_space_at_eol=True,
    )
    return MCPToolOutput(
        text=diff,
        max_chars=deps.settings().default_git_diff_response_max_chars,
    ).render()


@mcp.tool()
@log_inputs_outputs()
def add_tasks(
    tasks: Annotated[
        list[str],
        Field(
            min_length=1,
            max_length=10,
            description="Each string in the list is an individual task.",
        ),
    ],
    common_prefix: str | None = None,
) -> ToolResponse:
    """Add tasks to be completed by an LLM in the future.

    Do not add a task unless explicitly instructed to do so.

    When adding tasks, provide all required context.
    For example: step 1 set up the ~/git/p1 and ~/git/p2 repos projects step 2 load the diff with
        ref develop step 3 confirm that the function added in /examples/script.py is
        consistent with the existing /examples/other_script.py file.
    The common_prefix is prefixed to each task's action (if not None), it's provided
    to avoid having to repeat the common context for each task.

    Call this tool multiple times to add many tasks.
    """
    if common_prefix is not None:
        tasks = [f"{common_prefix} {action}" for action in tasks]

    with db.get_task_manager() as task_manager:
        for task_action in tasks:
            task_manager.add_task(task_action)
    return MCPToolOutput(
        text="Tasks(s) added. Do not immediately get a task unless explicitly told to do so.",
    ).render()


@mcp.tool()
@log_inputs_outputs()
def get_task() -> ToolResponse:
    """Get a single task.

    Do not use this tool unless explicitly told to do so.
    After you complete the task, mark it as done by calling the `set_task_done` tool.
    """
    with db.get_task_manager() as task_manager:
        db_task = task_manager.get_task()
        if db_task is None:
            return MCPToolOutput(text="no tasks").render()
        return MCPToolOutput(
            jsonable={
                "action": db_task.action,
                "id": db_task.id,
            },
        ).render()


@mcp.tool()
@log_inputs_outputs()
def mark_task_done(
    task_id: int,
    outcome: str,
    # Type hinting with `db.TaskFollowUpCriticality` directly seems to completely FRY
    # claude desktop - seems the schema it generates perhaps causes it to generate
    # invalid input data and crashes (?) the desktop app before it even sends any
    # data. So reproduce it as a Literal 🐀
    follow_up_criticality: Annotated[
        Literal["no_followup", "low", "medium", "high"],
        Field(description="If the task requires no follow up, set to no_followup"),
    ],
) -> ToolResponse:
    """Set a task as done wth a specific outcome.

    You can call this multiple times to update the outcome.
    """
    allowed_vals = {str(x) for x in db.TaskFollowUpCriticality.__members__.values()}
    if follow_up_criticality not in allowed_vals:
        raise ValueError(f"{follow_up_criticality} must be in {allowed_vals}")
    with db.get_task_manager() as task_manager:
        task_manager.set_task_done(
            task_id,
            outcome,
            db.TaskFollowUpCriticality(follow_up_criticality),
        )
    return MCPToolOutput(text="ok").render()


def _get_project_or_error(project_name: str) -> ToolProject:
    if project_name not in PROJECTS:
        raise ValueError(
            f"Project {project_name} not configured. Either double check the project name "
            f"or run the tool to set up a new project. The server may have been restarted "
            f"causing it to no longer be configured.",
        )
    return PROJECTS[project_name]


def _list_chunks_in_file(
    proj_file: ProjectFile,
    filter_: FilterType,
    filter_on: Literal["name", "name_or_content"],
) -> MCPToolOutput:
    target_file = proj_file.file
    chunks = [x for x in target_file.chunks if x.matches_filter(filter_, filter_on)]
    resp_data = [f"id={x.id_} (category={x.category} chars={len(x.content)})" for x in chunks]
    resp_text = "\n".join(resp_data)
    chunk_info = f"({len(chunks)} of {len(target_file.chunks)} chunks)"
    return MCPToolOutput(text=f"{chunk_info}\n{resp_text}")


def _filter_files_by_chunk(
    project_name: str,
    filter_: FilterType,
    filter_on: Literal["name", "name_or_content"],
) -> MCPToolOutput:
    project = _get_project_or_error(project_name)
    matching_files: set[pathlib.Path] = set()
    for file in project.chunk_project.files:
        if any(c.matches_filter(filter_, filter_on) for c in file.chunks):
            matching_files.add(file.abs_path)
    data = create_file_tree(project_root=project.root, paths=matching_files)
    if data is None:
        return MCPToolOutput(text="No files found")
    elif isinstance(data, str):
        return MCPToolOutput(text=data)
    else:
        assert_never(data)


if __name__ == "__main__":
    import time

    # mark_task_done(task_id=10, outcome="ok", follow_up_criticality="low")
    t1 = time.monotonic()
    configure_project(
        root_path=pathlib.Path("~/git/mcpunk"),
        project_name="mcpunk",
    )
    t2 = time.monotonic()
    print(f"Configured project in {(t2 - t1) * 1000:.2f}ms")
    _proj = PROJECTS["mcpunk"].chunk_project
    print(len([f for f in _proj.files if f.ext == ".py"]), "files")
    print(sum(len(f.contents.splitlines()) for f in _proj.files if f.ext == ".py"), "lines")
    print(sum(len(f.contents) for f in _proj.files if f.ext == ".py"), "chars")
    find_files_by_chunk_content(
        project_name="mcpunk",
        chunk_contents_filter=["desktop"],
    )
    _list_chunks_in_file(
        proj_file=ProjectFile(
            project_name="mcpunk",
            rel_path=pathlib.Path("README.md"),
        ),
        filter_=None,
        filter_on="name",
    )
    chunk_details(
        project_name="mcpunk",
        rel_path=pathlib.Path("README.md"),
        chunk_id="xxx",
    )
    # f = [
    #     x
    #     for x in PROJECTS["mcpunk"].chunk_project.files
    #     if x.abs_path == pathlib.Path(PROJECTS["mcpunk"].root / "docs/infrastructure.md")
    # ][0]
    diff_with_ref(
        project_name="mcpunk",
        ref="main",
    )



import json
from datetime import UTC, datetime, timedelta

import mcp.types as mcp_types
import pytest
import sqlalchemy as sa

from mcpunk import db, tools
from mcpunk.dependencies import deps
from mcpunk.settings import Settings


def test_add_tasks() -> None:
    """Test that add_tasks creates tasks with correct properties."""
    actions = ["task one", "task two"]
    prefix = "common stuff"
    tools.add_tasks(tasks=actions, common_prefix=prefix)

    with deps.session_maker().begin() as sess:
        tasks = sess.scalars(sa.select(db.Task).order_by(db.Task.id)).all()
        assert len(tasks) == 2
        for task, action in zip(tasks, actions, strict=True):
            assert task.action == f"{prefix} {action}"
            assert task.state == db.TaskState.TODO
            assert task.outcome is None
            assert task.last_picked_up_at is None
            assert task.follow_up_criticality is None


def test_get_task() -> None:
    """Test that get_task returns and updates task correctly."""
    # Create tasks
    tools.add_tasks(tasks=["task one", "task two"])

    # Get first task
    task_info_raw = tools.get_task()
    task_info_dict = json.loads(_squash_resp(task_info_raw))
    assert task_info_dict["action"] == "task one"
    task_id = task_info_dict["id"]

    # Verify task state updated
    with deps.session_maker().begin() as sess:
        task = sess.get_one(db.Task, task_id)
        assert task.state == db.TaskState.DOING
        assert task.last_picked_up_at is not None
        # Second task still in to do state
        other_task = sess.execute(sa.select(db.Task).where(db.Task.id != task_id)).scalar_one()
        assert other_task.state == db.TaskState.TODO


def test_get_task_none_available() -> None:
    """Test get_task when no tasks available."""
    result = tools.get_task()
    result_text = _squash_resp(result)
    assert result_text == "no tasks"


def test_mark_task_done() -> None:
    """Test marking a task as done with an outcome."""
    tools.add_tasks(tasks=["task one"])

    with deps.session_maker().begin() as sess:
        task_id = sess.scalars(sa.select(db.Task)).one().id

    outcome = "completed successfully"
    result = tools.mark_task_done(
        task_id=task_id,
        outcome=outcome,
        follow_up_criticality="no_followup",
    )
    result_text = _squash_resp(result)
    assert result_text == "ok"

    with deps.session_maker().begin() as sess:
        task = sess.get_one(db.Task, task_id)
        assert task.state == db.TaskState.DONE
        assert task.outcome == outcome
        assert task.follow_up_criticality == db.TaskFollowUpCriticality.NO_FOLLOWUP


def test_mark_task_done_invalid_id() -> None:
    """Test marking a non-existent task as done."""
    invalid_id = 12345
    with pytest.raises(ValueError, match=f"No task found with id {invalid_id}"):
        tools.mark_task_done(
            task_id=invalid_id,
            outcome="doesn't matter",
            follow_up_criticality="no_followup",
        )


def _squash_resp(resp: tools.ToolResponse) -> str:
    if isinstance(resp, mcp_types.TextContent):
        return resp.text
    elif isinstance(resp, list):
        if not all(isinstance(x, mcp_types.TextContent) for x in resp):
            raise NotImplementedError(f"Unexpected type: {type(resp)}")
        squashed = " ".join(x.text for x in resp)
        return squashed
    else:
        raise NotImplementedError(f"Unexpected type: {type(resp)}")


def test_get_task_visibility_timeout() -> None:
    """Test that tasks become visible again after timeout."""
    # We want to ensure it's a large number so that we don't fly past it and cause
    # a flaky test.
    settings = Settings(task_queue_visibility_timeout_seconds=10_000)

    with deps.override(settings_partial=settings):
        tools.add_tasks(tasks=["task one"])

        # Pick up task
        task_info_raw = tools.get_task()
        task_info_dict = json.loads(_squash_resp(task_info_raw))
        task_id = task_info_dict["id"]

        # Verify task unavailable immediately
        no_task_info = tools.get_task()
        assert _squash_resp(no_task_info) == "no tasks"

        # Move task pickup time to before timeout
        with deps.session_maker().begin() as sess:
            task = sess.get_one(db.Task, task_id)
            task.last_picked_up_at = (
                datetime.now(UTC)
                - deps.settings().task_queue_visibility_timeout
                - timedelta(seconds=1)
            )

        # Task should be available again
        task_info_raw = tools.get_task()
        task_info_dict = json.loads(_squash_resp(task_info_raw))
        assert task_info_dict["id"] == task_id  # Same task picked up



import pytest
from deepdiff import DeepDiff  # type: ignore[attr-defined]

from mcpunk.python_file_analysis import Callable, extract_imports, extract_module_statements

# This makes pytest print nice diffs when asserts fail
# within this function.
pytest.register_assert_rewrite("assert_calls_equal")


def sorted_callables(calls: list[Callable]) -> list[Callable]:
    """Sort a list of Call objects by line and column."""
    return sorted(calls, key=lambda x: (x.line, x.col, x.code))


def assert_callables_equal(actual: list[Callable], expected: list[Callable]) -> None:
    """Assert that two lists of Call objects are equal."""
    actual = sorted_callables(actual)
    expected = sorted_callables(expected)
    assert len(actual) == len(expected)
    actual_dicts = [x.model_dump(mode="json") for x in actual]
    expected_dicts = [x.model_dump(mode="json") for x in expected]
    if actual_dicts != expected_dicts:
        print("\n")  # Help format with pytest
        for i, (a, e) in enumerate(zip(actual_dicts, expected_dicts, strict=True)):
            if a != e:
                print(f"Expected {i}: {e}")
                print(f"Actual   {i}: {a}")
                print(f"Diff {i}: {DeepDiff(a, e)}")
            else:
                print(f"Pair {i} is equal")
        # raise AssertionError(DeepDiff(actual_dicts, expected_dicts))
    assert actual == expected


def test_empty_source_code() -> None:
    source = """\


    """
    actual = Callable.from_source_code(source)
    expected: list[Callable] = []
    assert_callables_equal(actual, expected)


def test_no_function_definitions() -> None:
    source = """\
x = 1
y = x + 2
    """
    actual = Callable.from_source_code(source)
    expected: list[Callable] = []
    assert_callables_equal(actual, expected)


def test_basic_function_def() -> None:
    source = """\
def my_func(x):
    return x
my_func("hello")

def empty_func():
    pass

def empty_func_ii():
    '''some docstring'''
    """
    actual = Callable.from_source_code(source)
    expected = [
        Callable(
            name="my_func",
            line=1,
            col=4,
            offset=4,
            code_offset_start=0,
            code_offset_end=28,
            code="def my_func(x):\n    return x",
        ),
        Callable(
            name="empty_func",
            line=5,
            col=4,
            offset=51,
            code_offset_start=47,
            code_offset_end=73,
            code="def empty_func():\n    pass",
        ),
        Callable(
            name="empty_func_ii",
            line=8,
            col=4,
            offset=79,
            code_offset_start=75,
            code_offset_end=75 + 45,
            code="def empty_func_ii():\n    '''some docstring'''",
        ),
    ]
    assert_callables_equal(actual, expected)


def test_function_def_with_callable() -> None:
    source = """\
@wrapper()
def my_func(x):
    return x
my_func("hello")
    """
    actual = Callable.from_source_code(source)
    expected = [
        Callable(
            name="my_func",
            line=2,
            col=4,
            offset=15,
            code_offset_start=0,
            code_offset_end=39,
            code="@wrapper()\ndef my_func(x):\n    return x",
        ),
    ]
    assert_callables_equal(actual, expected)


def test_simple_class() -> None:
    """Test basic class definition without methods."""
    source = """\
class SimpleClass:
    pass
    """
    actual = Callable.from_source_code(source)
    expected = [
        Callable(
            name="SimpleClass",
            line=1,
            col=6,
            offset=6,
            code_offset_start=0,
            code_offset_end=27,
            code="class SimpleClass:\n    pass",
        ),
    ]
    assert_callables_equal(actual, expected)


def test_async_function() -> None:
    """Test async function definition."""
    source = """\
async def fetch_data():
    return "data"
    """
    actual = Callable.from_source_code(source)
    expected = [
        Callable(
            name="fetch_data",
            line=1,
            col=10,
            offset=10,
            code_offset_start=0,
            code_offset_end=41,
            code='async def fetch_data():\n    return "data"',
        ),
    ]
    assert_callables_equal(actual, expected)


def test_nested_function() -> None:
    """Test nested function definition."""
    source = """\
def outer():
    def inner():
        return True
    return inner
    """
    actual = Callable.from_source_code(source)
    expected = [
        Callable(
            name="outer",
            line=1,
            col=4,
            offset=4,
            code_offset_start=0,
            code_offset_end=66,
            code="def outer():\n    def inner():\n        return True\n    return inner",
        ),
        Callable(
            name="inner",
            line=2,
            col=8,
            offset=21,
            code_offset_start=13,
            code_offset_end=13 + 36,
            code="    def inner():\n        return True",
        ),
    ]
    assert_callables_equal(actual, expected)


def test_class_with_method() -> None:
    """Test class with a single method."""
    source = """\
class MyClass:
    def my_method(self):
        return 42
    """
    actual = Callable.from_source_code(source)
    expected = [
        Callable(
            name="MyClass",
            line=1,
            col=6,
            offset=6,
            code_offset_start=0,
            code_offset_end=57,
            code="class MyClass:\n    def my_method(self):\n        return 42",
        ),
        Callable(
            name="my_method",
            line=2,
            col=8,
            offset=23,
            code_offset_start=15,
            code_offset_end=15 + 42,
            code="    def my_method(self):\n        return 42",
        ),
    ]
    assert_callables_equal(actual, expected)


def test_nested_classes() -> None:
    """Test nested class definitions with methods."""
    source = """\
class Outer:
    class Inner:
        def inner_method(self):
            return "inner"

        class DeepNested:
            @staticmethod
            def deep_method():
                return "deep"

    def outer_method(self):
        return "outer"
    """
    actual = Callable.from_source_code(source)
    expected = [
        Callable(
            name="Outer",
            line=1,
            col=6,
            offset=6,
            code_offset_start=0,
            code_offset_end=254,
            code="class Outer:\n"
            "    class Inner:\n"
            "        def inner_method(self):\n"
            '            return "inner"\n'
            "\n"
            "        class DeepNested:\n"
            "            @staticmethod\n"
            "            def deep_method():\n"
            '                return "deep"\n'
            "\n"
            "    def outer_method(self):\n"
            '        return "outer"',
        ),
        Callable(
            name="Inner",
            line=2,
            col=10,
            offset=23,
            code_offset_start=13,
            code_offset_end=13 + 189,
            code="    class Inner:\n"
            "        def inner_method(self):\n"
            '            return "inner"\n'
            "\n"
            "        class DeepNested:\n"
            "            @staticmethod\n"
            "            def deep_method():\n"
            '                return "deep"',
        ),
        Callable(
            name="inner_method",
            line=3,
            col=12,
            offset=42,
            code_offset_start=30,
            code_offset_end=30 + 58,
            code='        def inner_method(self):\n            return "inner"',
        ),
        Callable(
            name="DeepNested",
            line=6,
            col=14,
            offset=104,
            code_offset_start=90,
            code_offset_end=90 + 112,
            code="        class DeepNested:\n"
            "            @staticmethod\n"
            "            def deep_method():\n"
            '                return "deep"',
        ),
        Callable(
            name="deep_method",
            line=8,
            col=16,
            offset=158,
            code_offset_start=116,
            code_offset_end=116 + 86,
            code="            @staticmethod\n"
            "            def deep_method():\n"
            '                return "deep"',
        ),
        Callable(
            name="outer_method",
            line=11,
            col=8,
            offset=212,
            code_offset_start=204,
            code_offset_end=204 + 50,
            code='    def outer_method(self):\n        return "outer"',
        ),
    ]
    assert_callables_equal(actual, expected)


def test_multiline_function_definition() -> None:
    """Test function definitions that span multiple lines."""
    source = """\
def long_function_name(
    param1: str,
    param2: int,
    param3: list[
        dict[str, int]
    ],
    *args: tuple[str, ...],
    **kwargs: dict[
        str,
        Any
    ],
) -> None:
    return None
    """
    actual = Callable.from_source_code(source)
    expected = [
        Callable(
            name="long_function_name",
            line=1,
            col=4,
            offset=4,
            code_offset_start=0,
            code_offset_end=212,
            code="def long_function_name(\n"
            "    param1: str,\n"
            "    param2: int,\n"
            "    param3: list[\n"
            "        dict[str, int]\n"
            "    ],\n"
            "    *args: tuple[str, ...],\n"
            "    **kwargs: dict[\n"
            "        str,\n"
            "        Any\n"
            "    ],\n"
            ") -> None:\n"
            "    return None",
        ),
    ]
    assert_callables_equal(actual, expected)


def test_multiline_class_definition() -> None:
    """Test class definitions that span multiple lines."""
    source = """\
class ComplexClass(
    BaseClass,
    Generic[
        TypeVar1,
        TypeVar2,
    ],
    metaclass=ABCMeta,
):
    '''A complex class with multi-line definition'''

    def __init__(
        self,
        param1: str,
        param2: int = 42
    ) -> None:
        self.param1 = param1
        self.param2 = param2
    """
    actual = Callable.from_source_code(source)
    expected = [
        Callable(
            name="ComplexClass",
            line=1,
            col=6,
            offset=6,
            code_offset_start=0,
            code_offset_end=321,
            code="class ComplexClass(\n"
            "    BaseClass,\n"
            "    Generic[\n"
            "        TypeVar1,\n"
            "        TypeVar2,\n"
            "    ],\n"
            "    metaclass=ABCMeta,\n"
            "):\n"
            "    '''A complex class with multi-line definition'''\n"
            "\n"
            "    def __init__(\n"
            "        self,\n"
            "        param1: str,\n"
            "        param2: int = 42\n"
            "    ) -> None:\n"
            "        self.param1 = param1\n"
            "        self.param2 = param2",
        ),
        Callable(
            name="__init__",
            line=11,
            col=8,
            offset=179,
            code_offset_start=171,
            code_offset_end=171 + 150,
            code="    def __init__(\n"
            "        self,\n"
            "        param1: str,\n"
            "        param2: int = 42\n"
            "    ) -> None:\n"
            "        self.param1 = param1\n"
            "        self.param2 = param2",
        ),
    ]
    assert_callables_equal(actual, expected)


def test_multiline_decorators() -> None:
    """Test function with multi-line decorators."""
    source = """\
@decorator1(
    param1="value1",
    param2="value2",
)
@decorator2(
    param3={
        "key1": "value1",
        "key2": "value2",
    }
)
def decorated_function(
    x: int,
    y: str,
) -> bool:
    return True
    """
    actual = Callable.from_source_code(source)
    expected = [
        Callable(
            name="decorated_function",
            line=11,
            col=4,
            offset=147,
            code_offset_start=0,
            code_offset_end=217,
            code="@decorator1(\n"
            '    param1="value1",\n'
            '    param2="value2",\n'
            ")\n"
            "@decorator2(\n"
            "    param3={\n"
            '        "key1": "value1",\n'
            '        "key2": "value2",\n'
            "    }\n"
            ")\n"
            "def decorated_function(\n"
            "    x: int,\n"
            "    y: str,\n"
            ") -> bool:\n"
            "    return True",
        ),
    ]
    assert_callables_equal(actual, expected)


def test_unicode_names() -> None:
    """Test handling of Unicode characters in function and class names."""
    source = """\
def 你好_world():
    pass

class お早う_class:
    def μ_method():
        pass

def λ_function():
    yield from range(10)
    """
    actual = Callable.from_source_code(source)
    expected = [
        Callable(
            name="你好_world",
            line=1,
            col=4,
            offset=4,
            code_offset_start=0,
            code_offset_end=24,
            code="def 你好_world():\n    pass",
        ),
        Callable(
            name="お早う_class",
            line=4,
            col=6,
            offset=32,
            code_offset_start=26,
            code_offset_end=26 + 49,
            code="class お早う_class:\n    def μ_method():\n        pass",
        ),
        Callable(
            name="μ_method",
            line=5,
            col=8,
            offset=51,
            code_offset_start=43,
            code_offset_end=43 + 32,
            code="    def μ_method():\n        pass",
        ),
        Callable(
            name="λ_function",
            line=8,
            col=4,
            offset=81,
            code_offset_start=77,
            code_offset_end=77 + 42,
            code="def λ_function():\n    yield from range(10)",
        ),
    ]
    assert_callables_equal(actual, expected)


def test_conditional_definitions() -> None:
    """Test functions and classes defined inside conditional blocks."""
    source = """\
if True:
    def true_func():
        pass
else:
    def false_func():
        pass

try:
    class TryClass:
        pass
except Exception:
    class ExceptClass:
        pass
finally:
    def finally_func():
        return True
    """
    actual = Callable.from_source_code(source)
    expected = [
        Callable(
            name="true_func",
            line=2,
            col=8,
            offset=17,
            code_offset_start=9,
            code_offset_end=9 + 33,
            code="    def true_func():\n        pass",
        ),
        Callable(
            name="false_func",
            line=5,
            col=8,
            offset=57,
            code_offset_start=49,
            code_offset_end=49 + 34,
            code="    def false_func():\n        pass",
        ),
        Callable(
            name="TryClass",
            line=9,
            col=10,
            offset=100,
            code_offset_start=90,
            code_offset_end=90 + 32,
            code="    class TryClass:\n        pass",
        ),
        Callable(
            name="ExceptClass",
            line=12,
            col=10,
            offset=151,
            code_offset_start=141,
            code_offset_end=141 + 35,
            code="    class ExceptClass:\n        pass",
        ),
        Callable(
            name="finally_func",
            line=15,
            col=8,
            offset=194,
            code_offset_start=186,
            code_offset_end=186 + 43,
            code="    def finally_func():\n        return True",
        ),
    ]
    assert_callables_equal(actual, expected)


def test_property_and_special_methods() -> None:
    """Test property decorators and special methods."""
    source = """\
class DataClass:
    @property
    def value(self) -> int:
        return self._value

    @value.setter
    def value(self, val: int) -> None:
        self._value = val

    @classmethod
    def from_string(cls, s: str) -> 'DataClass':
        return cls(int(s))

    @staticmethod
    def validate(x: int) -> bool:
        return x > 0

    def __str__(self) -> str:
        return str(self._value)

    async def __aiter__(self):
        yield self._value
    """
    actual = Callable.from_source_code(source)
    expected = [
        Callable(
            name="DataClass",
            line=1,
            col=6,
            offset=6,
            code_offset_start=0,
            code_offset_end=458,
            code="class DataClass:\n"
            "    @property\n"
            "    def value(self) -> int:\n"
            "        return self._value\n\n"
            "    @value.setter\n"
            "    def value(self, val: int) -> None:\n"
            "        self._value = val\n\n"
            "    @classmethod\n"
            "    def from_string(cls, s: str) -> 'DataClass':\n"
            "        return cls(int(s))\n\n"
            "    @staticmethod\n"
            "    def validate(x: int) -> bool:\n"
            "        return x > 0\n\n"
            "    def __str__(self) -> str:\n"
            "        return str(self._value)\n\n"
            "    async def __aiter__(self):\n"
            "        yield self._value",
        ),
        Callable(
            name="value",
            line=3,
            col=8,
            offset=39,
            code_offset_start=17,
            code_offset_end=17 + 68,
            code="    @property\n    def value(self) -> int:\n        return self._value",
        ),
        Callable(
            name="value",
            line=7,
            col=8,
            offset=113,
            code_offset_start=87,
            code_offset_end=87 + 82,
            code="    @value.setter\n    def value(self, val: int) -> None:\n"
            "        self._value = val",
        ),
        Callable(
            name="from_string",
            line=11,
            col=8,
            offset=196,
            code_offset_start=171,
            code_offset_end=171 + 92,
            code="    @classmethod\n    def from_string(cls, s: str) -> 'DataClass':\n"
            "        return cls(int(s))",
        ),
        Callable(
            name="validate",
            line=15,
            col=8,
            offset=291,
            code_offset_start=265,
            code_offset_end=265 + 72,
            code="    @staticmethod\n    def validate(x: int) -> bool:\n        return x > 0",
        ),
        Callable(
            name="__str__",
            line=18,
            col=8,
            offset=347,
            code_offset_start=339,
            code_offset_end=339 + 61,
            code="    def __str__(self) -> str:\n        return str(self._value)",
        ),
        Callable(
            name="__aiter__",
            line=21,
            col=14,
            offset=416,
            code_offset_start=402,
            code_offset_end=402 + 56,
            code="    async def __aiter__(self):\n        yield self._value",
        ),
    ]
    assert_callables_equal(actual, expected)


def test_extract_imports_basic() -> None:
    source = """\
import os
import sys as system
import json, csv
from pathlib import Path
from typing import List, Optional as Opt
from . import local_module
from ..parent import something as other
"""

    result = extract_imports(source)
    assert result == [
        "import os",
        "import sys as system",
        "import json, csv",
        "from pathlib import Path",
        "from typing import List, Optional as Opt",
        "from . import local_module",
        "from ..parent import something as other",
    ]


def test_extract_mixed_scope_imports() -> None:
    source = """\
import os  # module level

def func1():
    import sys  # inside function
    print("hello")

import json  # module level again

class MyClass:
    import csv  # inside class

    def method(self):
        from pathlib import Path  # inside method
        return Path('.')

from typing import List  # module level at end
"""

    result = extract_imports(source)
    assert result == ["import os", "import json", "from typing import List"]


def test_extract_module_statements() -> None:
    source = """\
import os
x = 1

def func1():
    y = 2
    return y

CONSTANT = "test"
import json

class MyClass:
    z = 3

    def method(self):
        pass

final_var = True

a = (
    1,
    # Internal comment
    2,
)

if __name__ == "__main__":
    print("hey!")

# A comment!

async def func2():
    pass
"""

    result = extract_module_statements(source)
    assert result == [
        "x = 1",
        "def func1...",
        'CONSTANT = "test"',
        "class MyClass...",
        "final_var = True",
        "a = (\n    1,\n    # Internal comment\n    2,\n)",
        'if __name__ == "__main__":\n    print("hey!")',
        "async def func2...",
    ]



from pathlib import Path

from mcpunk.file_chunkers import WholeFileChunker


def test_whole_file_chunker() -> None:
    source_code = "some\ncontent"

    chunks = WholeFileChunker(source_code, Path("test.txt")).chunk_file()

    assert len(chunks) == 1
    chunk = chunks[0]
    assert chunk.category.value == "whole_file"
    assert chunk.name == "<whole_file>"
    assert chunk.line == 1
    assert chunk.content == "some\ncontent"

    # Accepts any file
    assert WholeFileChunker.can_chunk("", Path("whatever.txt"))
    assert WholeFileChunker.can_chunk("", Path("test.xyz"))



import ast
import logging
from functools import lru_cache
from typing import Annotated, assert_never

import asttokens
import more_itertools
from asttokens.util import Token
from asttokens.util import walk as asttokens_walk
from pydantic import BaseModel, ConfigDict, Field

logger = logging.getLogger(__name__)


@lru_cache(maxsize=1024)
def _ast_cache(src: str) -> asttokens.ASTTokens:
    """Source code to asttokens AST with caching.

    Note that the cache is not particularly large, so to make effective use of it
    you probably want to things like `for file: for analysis: ast = _ast_cache(...)`
    rather than `for analysis: for file: ast = _ast_cache(...)`
    """
    return asttokens.ASTTokens(src, parse=True)


class Callable(BaseModel):
    """Represents a callable (function/class) definition in the source code."""

    name: Annotated[str, Field(description="Name of the function or class")]
    # line/col/offset are for where `my_func` begins in `def my_func(...):` or
    # `MyClass` in `class MyClass(...):`
    line: Annotated[int, Field(description="First line is 1")]
    col: Annotated[int, Field(description="First column is 0")]
    offset: Annotated[int, Field(description="char offset corresponding to line & col.")]
    code_offset_start: Annotated[
        int,
        Field(description="Character offset of the start of the callable"),
    ]
    code_offset_end: Annotated[
        int,
        Field(description="Character offset of the end of the callable"),
    ]
    # This is the whole code block. For a function it will include decorators,
    # for example. Does NOT line up with line/col/offset.
    code: Annotated[str, Field(description="Complete source code of the callable")]

    model_config = ConfigDict(validate_assignment=True)

    @classmethod
    def from_source_code(
        cls,
        source_code: str,
    ) -> list["Callable"]:
        """Extract all callables from the given source code."""
        try:
            atok = _ast_cache(source_code)
        except SyntaxError:
            logger.error(  # noqa: TRY400
                f"Skipping {source_code} because it's not valid Python syntax",
            )
            return []
        assert atok.tree is not None
        nodes: list[ast.FunctionDef | ast.ClassDef | ast.AsyncFunctionDef] = [
            n
            for n in asttokens_walk(
                atok.tree,
                # Joined strings seem to relate to f-strings, and enabling
                # them seems to be problematic. Let's just skip them.
                include_joined_str=False,
            )
            if isinstance(n, ast.FunctionDef | ast.ClassDef | ast.AsyncFunctionDef)
        ]
        callables = []
        for node in nodes:
            # print(ast.dump(node, indent=4))
            # print(atok.get_text(node))

            callable_name = node.name

            prefixes: list[str]
            if isinstance(node, ast.FunctionDef):
                prefixes = ["def"]
            elif isinstance(node, ast.AsyncFunctionDef):
                prefixes = ["async", "def"]
            elif isinstance(node, ast.ClassDef):
                prefixes = ["class"]
            else:
                assert_never(node)

            all_tokens: list[Token] = list(atok.get_tokens(node))
            # Make sure we sort by position in the file
            all_tokens = sorted(all_tokens, key=lambda x: x.startpos)

            callable_start_tok: Token | None = None
            window_size = len(prefixes) + 1
            looking_for = [*prefixes, callable_name]
            for window_elements in more_itertools.sliding_window(all_tokens, window_size):
                # This is a bit gross
                assert len(window_elements) == window_size
                if looking_for == [x.string for x in window_elements]:
                    callable_start_tok = window_elements[-1]
                    break

            if callable_start_tok is None:
                raise AssertionError(
                    f"Could not find callable {callable_name} in {node}\n"
                    f"{atok.get_text(node)}\n"
                    f"{ast.dump(node, indent=4)}\n",
                )

            range_ = atok.get_text_range(node)
            callable_ = Callable(
                name=node.name,
                line=callable_start_tok.start[0],
                col=callable_start_tok.start[1],
                offset=callable_start_tok.startpos,
                code=atok.get_text(node),
                code_offset_start=range_[0],
                code_offset_end=range_[1],
            )
            callables.append(callable_)
        return callables


def extract_imports(source_code: str) -> list[str]:
    """Extract all module-level import statements from source code.

    Takes source code as input and returns a list of import statements.
    """
    atok = _ast_cache(source_code)
    imports: list[str] = []

    for node in atok.tree.body:  # type: ignore[union-attr]
        if isinstance(node, ast.Import | ast.ImportFrom):
            import_code = atok.get_text(node)
            imports.append(import_code)

    return imports


def extract_module_statements(source_code: str) -> list[str]:
    """Extract all module-level statements from source code.

     function/class definitions are inserted like `def func1...` or `class MyClass...`
     to provide context around where the module-level statements are defined.

    Takes source code as input and returns a list of statement strings.
    """
    # TODO: include comments

    atok = _ast_cache(source_code)
    statements: list[str] = []

    for node in atok.tree.body:  # type: ignore[union-attr]
        if isinstance(node, ast.FunctionDef):
            statements.append(f"def {node.name}...")
        elif isinstance(node, ast.AsyncFunctionDef):
            statements.append(f"async def {node.name}...")
        elif isinstance(node, ast.ClassDef):
            statements.append(f"class {node.name}...")
        elif isinstance(node, ast.Import | ast.ImportFrom):
            pass
        else:
            statement_code = atok.get_text(node)
            statements.append(statement_code)

    return statements



from pathlib import Path

import pytest
import sqlalchemy as sa

from mcpunk import db
from mcpunk.dependencies import deps
from mcpunk.settings import Settings


def test_init_db_creates_new_db(tmp_path: Path) -> None:
    # There's an auto-use fixture that uses tmp_path to set up a fresh db
    # for each test, and already runs init. So we need to just fiddle the path
    # to ensure we really do get a fresh db.
    tmp_path = tmp_path / "make_it_fresher" / "even_fresher"

    settings = Settings(db_path=tmp_path / "test.db")
    assert not settings.db_path.absolute().exists()
    with deps.override(settings=settings):
        db.init_db()
        assert settings.db_path.exists()

        # Verify PRAGMA statements
        expected_pragmas = [
            ("PRAGMA journal_mode", "wal"),
            ("PRAGMA synchronous", "1"),  # 1 is normal
            ("PRAGMA busy_timeout", "5000"),
            ("PRAGMA cache_size;", "-20000"),
            ("PRAGMA foreign_keys", "1"),
        ]
        with deps.session_maker().begin() as sess:
            for stmt, expected_result in expected_pragmas:
                result = sess.execute(sa.text(stmt)).scalar()
                assert str(result) == expected_result

            # Verify version
            version = sess.scalars(sa.select(db.DBVersion.version)).one()
            assert version == db.CURRENT_DB_VERSION

            actual_tables = sess.execute(
                sa.text("SELECT name FROM sqlite_master WHERE type='table';"),
            ).fetchall()
            actual_tables = sorted(actual_tables)
            expected_tables = sorted([("db_version",), ("task",)])
            assert actual_tables == expected_tables


def test_init_db_version_mismatch() -> None:
    db.init_db()  # This will double check the version is okay right now

    # Modify version to be different
    with deps.session_maker().begin() as sess:
        sess.execute(sa.delete(db.DBVersion))
        sess.add(db.DBVersion(version=db.CURRENT_DB_VERSION + "abcdefsg"))

    with pytest.raises(ValueError, match="Database version mismatch"):
        db.init_db()
